"""
Python implementation of a Genetic Algorithm in a functional manner.
"""

import random
from typing import List, Optional

import numpy as np  # type: ignore

from mendelopt.common.types import FitnessFunc, GeneticParams, GeneticResult, Individual
from mendelopt.generation import GenerationStrategy
from mendelopt.mutation import MutationStrategy
from mendelopt.reproduction import ReproductionStrategy


def optimize(
    fitness: FitnessFunc,
    generation_strategy: GenerationStrategy,
    mutation_strategy: MutationStrategy,
    reproduction_strategy: ReproductionStrategy,
    params: GeneticParams = GeneticParams(),
    warm_start: Optional[List[Individual]] = None,
) -> GeneticResult:
    """
    Performs genetic optimization on a function that expected a NumPy array
    and returns a floating point value indicating that array's fitness.

    See GeneticParams for further details on the specific optimization parameters.

    Each evolution step is implemented as follows:
        1. The fittest individuals are carried over as per elite_rate.
        2. A proportion of individuals given by (1 - elite_rate) * (1 - crossover_rate),
           rounded up, will be generated as mutations from fit individuals¹.
        3. A proportion of individuals given by (1 - elite_rate) * crossover_rate,
           rounded down, will be generated as crossovers from fit individuals¹.

    ¹The individuals selected for the mutation and crossover steps are
     randomly selected with probability given by their relative fitness
     compared to the current population.
    """

    fitness_history = np.empty(params.evolution_steps)

    population = sorted(
        warm_start
        if warm_start is not None
        else generation_strategy.generate_population(params.population_size, fitness),
        key=lambda individual: -individual.fitness(),
    )

    for iteration_count in range(params.evolution_steps):
        elite = population[: params.elite_count]

        immigrants = generation_strategy.generate_population(params.immigrant_count, fitness)

        mutation_origins = _select_candidates(population, params.mutation_count)
        mutations = [mutation_strategy.mutate(individual, fitness) for individual in mutation_origins]

        parents_a = _select_candidates(population, params.newborn_count)
        parents_b = _select_candidates(population, params.newborn_count)
        newborns = [reproduction_strategy.reproduce(parents, fitness) for parents in zip(parents_a, parents_b)]

        new_population = elite
        new_population.extend(immigrants)
        new_population.extend(mutations)
        new_population.extend(newborns)

        population = sorted(new_population, key=lambda individual: -individual.fitness())
        fitness_history[iteration_count] = population[0].fitness()

    return GeneticResult(
        best=population[0],
        history=fitness_history,
        population=population,
    )


def _select_candidates(population: List[Individual], count: int) -> List[Individual]:
    """
    Select candidates for an evolution step based on fitness.
    """

    return random.choices(population, weights=[individual.fitness() for individual in population], k=count)



"""
Implements abstract and basic individual mutation strategies.
"""

import random
from abc import ABC, abstractmethod
from typing import List, Set

from mendelopt.common.types import FitnessFunc, Individual
from mendelopt.common.utils import censor_string


class MutationStrategy(ABC):
    """
    Abstract class defining abstract methods for an individual mutation strategy.
    """

    @abstractmethod
    def mutate(self, individual: Individual, fitness_function: FitnessFunc) -> Individual:
        """
        Function to mutate an individual.
        """


class BasicMutationStrategy(MutationStrategy):
    """
    Basic mutation strategy based on randomly altering a single entry.
    """

    _domain: List[int]

    def __init__(self, values: Set[int]):
        self._domain = list(values)

    def mutate(self, individual: Individual, fitness_function: FitnessFunc) -> Individual:
        """
        Mutates an individual by switching one of its entries to another value.
        """

        mutation_index = tuple(random.randrange(0, dim_size) for dim_size in individual.as_array().shape)

        new_array = individual.as_array().copy()
        new_array[mutation_index] = random.sample(self._domain, k=1)[0]
        return Individual(new_array, fitness_function)

    def __repr__(self) -> str:
        domain_str = censor_string(str(self._domain))
        return f"BasicMutationStrategy(over {domain_str})"



"""
Implements abstract and basic individual generation strategies.
"""

import random
from abc import ABC, abstractmethod
from functools import reduce
from typing import List, Set, Tuple

import numpy as np  # type: ignore

from mendelopt.common.types import FitnessFunc, Individual
from mendelopt.common.utils import censor_string


class GenerationStrategy(ABC):
    """
    Abstract class defining abstract methods for an individual generation
    strategy.
    """

    @abstractmethod
    def generate_individual(self, fitness_function: FitnessFunc) -> Individual:
        """
        Function to generate an individual.
        """

    def generate_population(self, count: int, fitness_function: FitnessFunc) -> List[Individual]:
        """
        Function to generate a population of individuals.
        """

        return [self.generate_individual(fitness_function) for _ in range(count)]


class BasicGenerationStrategy(GenerationStrategy):
    """
    Basic generation strategy based on an array shape and integer domain
    values.
    """

    _element_count: int
    _shape: Tuple[int, ...]
    _domain: List[int]

    def __init__(self, shape: Tuple[int, ...], values: Set[int]):
        self._element_count = reduce(lambda x, y: x * y, shape)
        self._shape = shape
        self._domain = list(values)

    def generate_individual(self, fitness_function: FitnessFunc) -> Individual:
        """
        Generates a single random individual.
        """

        values = random.choices(self._domain, k=self._element_count)
        individual_array = np.reshape(values, self._shape)
        return Individual(individual_array, fitness_function)

    def __repr__(self) -> str:
        shape_str = censor_string(str(self._shape))
        domain_str = censor_string(str(self._domain))
        return f"BasicGenerationStrategy(shape {shape_str} over {domain_str})"



"""
Implements abstract and basic individual reproduction strategies.
"""

import random
from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np  # type: ignore

from mendelopt.common.types import FitnessFunc, Individual


class ReproductionStrategy(ABC):
    """
    Abstract class defining abstract methods for reproduction strategy.
    """

    @abstractmethod
    def reproduce(self, parents: Tuple[Individual, Individual], fitness_function: FitnessFunc) -> Individual:
        """
        Function to create a newborn from a couple of individuals.
        """


class BasicReproductionStrategy(ReproductionStrategy):
    """
    Basic reproduction strategy based on slicing.
    """

    def reproduce(self, parents: Tuple[Individual, Individual], fitness_function: FitnessFunc) -> Individual:
        """
        Creates a new individual based on slicing the parents.
        """

        new_array = np.empty(parents[0].as_array().size)

        slice_index = random.randint(0, new_array.size)
        if slice_index == 0:
            new_array[:] = parents[0].as_array().reshape(new_array.size)
        elif slice_index == new_array.size:
            new_array[:] = parents[1].as_array().reshape(new_array.size)
        else:
            new_array[:slice_index] = parents[0].as_array().reshape(new_array.size)[:slice_index]
            new_array[slice_index:] = parents[1].as_array().reshape(new_array.size)[slice_index:]

        return Individual(new_array.reshape(parents[0].as_array().shape), fitness_function)

    def __repr__(self) -> str:
        return "BasicReproductionStrategy()"



#!/usr/bin/env python

import setuptools


setuptools.setup(
    name="mendel-opt",
    packages=["mendelopt"],
    version="0.1",
    license="MIT",
    description="Python library written as a case study in implementing pythonic Genetic Algorithms.",
    author="Victhor Sartório",
    author_email="victhor@vsartor.com",
    url="https://github.com/vsartor/mendel-opt",
    keywords=["optimizer", "genetic algorithm"],
    install_requires=[
        "numpy",
    ],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Typing :: Typed",
    ],
)
